/* domme-test-driver -- some sanity checks for domme
Copyright (C) 2023, 2024 phantom (phantom@syslbnth.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>. */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <stdbool.h>
#include <sysexits.h>
#include <err.h>
#include <sys/ucontext.h>
/* Remember,
 * we want to test a header that may not have been installed yet. */
#include "domme.h"

extern const char *program_invocation_short_name;

/* Where are we in the test? */
enum test_phase { BUFFER = 1, ACT, OLDACT, VERSION, MSGTBL, XLAT_A, HANDLER } phase;

/* If you think you know why I chose this number, email me. */
const int CRASH_ARG = 8426;	// hint: 8 Apr 1926

/* Index variable for loops in $main,
   made global so it can be accessed from the segfault handler below. */
unsigned int i = 0;

void
handler(int sig, siginfo_t *info, void *ucontext) {
	/* $ucontext has to be declared as a pointer to void,
	 * but also needs to be dereferenced later. */
	ucontext_t *context = ucontext;

	/* We enter this function at most twice. */
	static bool re_entry = false;
	
	switch (phase) {
	case ACT:
		printf("ACT=ERROR\n");
		exit(phase);
	case OLDACT:
		printf("OLDACT=ERROR\n");
		exit(phase);
	case VERSION:
		/* We got here from the middle of a $printf call,
		 * but we should assume $printf won't output anything
		 * until its format string has been fully processed.
		 * So we have to reprint the version line from the beginning. */
		printf("VERSION=ERROR\n");
		exit(phase);
	case MSGTBL:
		printf("MSGTBL=ERROR:%d\n", i);
		exit(phase);
	case XLAT_A:
		printf("XLAT_A=ERROR:%d\n", i);
		exit(phase);
	case HANDLER:
		if (!re_entry) {
			re_entry = true;
			fprintf(stderr, "Attempting re-entry to the handler...\n");
			return;
		}

		printf("RE_ENTRY=OK\n");
		
		/* The next few checks are meant to make sure
		 * that none of our arguments were damaged in transit.
		 * The exact values of the locations checked aren't that important,
		 * only whether they are where we expect them to be. */
		if (!(sig == SIGSEGV
		      && info
		      && info->si_signo == SIGSEGV
		      && info->si_code  == SEGV_MAPERR
		      && info->si_addr  == CRASH	// this should be where we crashed
		      && context
		      /* The most important test. Also the scariest. */
		      && context->uc_mcontext.gregs[REG_RDI] == CRASH_ARG)) {
			printf("HANDLER=ERROR\n");
			exit(phase);
		}
		
		printf("HANDLER=OK\n");

		/* Finally, try using $domme with default behavior.
		 * Because we're already in a signal handler,
		 * we have to manually unblock SIGSEGV,
		 * which was blocked by the kernel on our arrival. */
		fprintf(stderr, "\nTriggering segfault without handler...\n");
		sigset_t set = { 0 };
		sigemptyset(&set);
		sigaddset(&set, SIGSEGV);
		sigprocmask(SIG_UNBLOCK, &set, NULL);
		
		/* If this program exits normally,
		 * your computer is haunted. */
		DOMME_init(NULL, NULL, sizeof(sigset_t)); // test NULL behavior
		*(char **)NULL = "Hi GDB!";
		__attribute__ ((fallthrough));
	default:		// You should never see this.
		fprintf(stderr,
			"%s: Internal error in test program (check exit code)\n",
			program_invocation_short_name);
		exit(phase);	// phase == bad if unknown, even worse if HANDLER
	}
}

/* This should never get called. */
void
wrong_handler () {
	fprintf(stderr, "Bad news--$rt_sigaction writes $oldact before reading $act.\n");
	exit(HANDLER+1);
}

int
main (int argc, char *argv[]) {
	/* To test how $oldact is saved,
	 * we need something to be registered before our real handler. */
	struct sigaction fake_sa = { 0 };
	fake_sa.sa_flags = SA_SIGINFO;
	fake_sa.sa_sigaction = wrong_handler;
	sigaction(SIGSEGV, &fake_sa, NULL);
	
	struct sigaction sa = { 0 };
	sa.sa_flags = SA_SIGINFO;
	sa.sa_sigaction = handler;
	if (argc > 1) {
		if (!strcmp(argv[1], "--short-circuit"))
			sa.sa_sigaction = (void *)domme;	// $DOMME_init should catch this
		else	// there's at least one unknown argument
			errx(EX_USAGE, "Unknown argument detected");
	}
	
	DOMME_init(&sa, &sa, sizeof(sigset_t));

	/* An error code higher than 63 denotes an error in the test program itself. */
	if (fprintf(stderr, "Now testing libdomme.\n") < 0
	    /* An error in that last write might not have been caught
	     * if its file was block-buffered. Better make sure. */
	    || fflush(stderr) == EOF)
		exit(EX_IOERR); 	// oh come on

	if (printf("OUTPUT=OK\n") < 0 || fflush(stdout) == EOF)
		exit(EX_IOERR);

	/* Before we can compare $sa and $fake_sa,
	 * we have to undo whatever libc might've done to $fake_sa
	 * between buffering it and calling $rt_sigaction. */
	phase = BUFFER;
	sa.sa_flags = sa.sa_flags &~ 0x04000000; // (SA_RESTORER)
	sa.sa_restorer = NULL;
	if (memcmp(&sa, &fake_sa, sizeof(sa))) {
		printf("BUFFER=ERROR\n");
		exit(phase);
	}
	printf("BUFFER=OK\n");

	/* Does $DOMME_init catch bad addresses
	 * on behalf of $rt_sigaction?
	 * If not, one of the calls below will segault. */
	phase = ACT;
	DOMME_init((void *)-1L, (void *)-1L, 0);
	printf("ACT=OK\n");
	phase = OLDACT;
	DOMME_init(0, (void *)-1L, 0); /* NULL should be fine, though */
	printf("OLDACT=OK\n");

	/* Let's see if the globals in "domme.h" have been properly defined.
	 * Be aware that this tests only whether they point to valid memory.
	 * If they do, errors in the printed output are rare,
	 * but you should still keep an eye out for them. */
	
	/* Print version number.
	 * Also, verify $stdout is legit, as with $stderr earlier. */
	phase = VERSION;
	printf("VERSION=%d.%d.%d.%d\n",
	       DOMME_version[0],
	       DOMME_version[1],
	       DOMME_version[2],
	       DOMME_version[3]);

	/* Try printing the script. */
	phase = MSGTBL;
	for (i = 0; i < DOMME_MSGTBL_SIZE; ++i)
		fprintf(stderr, "%s\n", DOMME_msgtbl[i].msgptr);
	printf("MSGTBL=OK\n");

	/* Try printing the XLAT array. */
	phase = XLAT_A;
	for (i = 0; i < DOMME_XLAT_A_SIZE; ++i)
		fprintf(stderr,
			"%3.2x%s",
			DOMME_xlat_a[i], (i & 15) == 15 ? "\n" : "");
	printf("XLAT_A=OK\n");

	/* So far, so good. Trigger a segfault and see what happens.
	 * We give our crash a (thematically appropriate) numerical argument
	 * so that we can later check whether it appears in the right place
	 * in the saved context passed to the handler.
	 * Because this was the first (and only) argument,
	 * it should be placed in %rdi, according to the SysV AMD64 ABI. */
	fprintf(stderr, "\nTriggering segfault with handler...\n");
	phase = HANDLER;
	CRASH(CRASH_ARG);

	/* Can't happen.
	 *
	 * The handler should drop us right back at the call to $CRASH
	 * after exiting once, and we should never reach this point--
	 * That is, per POSIX signal handling, execution should resume
	 * _at_ the offending instruction, not _after_ it.
	 * If you somehow you get here anyway,
	 * you might want to save any open files now,
	 * because something is very wrong with your system,
	 * and you are going to find out what very soon. */
	errx(EX_OSERR, "Handler re-entry failed");
}
